# the semantic version scheme here is read as follows:
# The major version: Only incremented on major breaking changes
# The minor version: Incremented upon each minor release.
# The patch version: Corresponds to Trino's grammar version.
__version__ = "0.0.368"
